import { id as mid } from "module";
import { AccumulatorFn, FusedProducer, Observable, Observer, OperatorFn } from "./interfaces";
import { TraceSource } from "@implab/core-amd/log/TraceSource";
import { scan } from "dojo/parser";
import { filter } from "dojo/query";

const trace = TraceSource.get(mid);

const noop = () => { };

const errorFallback = (e: unknown) => trace.error("Unhandled observable error: {0}", e);

const sink = <T>(consumer: Observer<T>) => {
    // eslint-disable-next-line @typescript-eslint/unbound-method
    const { next, error, complete } = consumer;
    return {
        next: next ? next.bind(consumer) : noop,
        error: error ? error.bind(consumer) : errorFallback, // report unhandled errors
        complete: complete ? complete.bind(consumer) : noop
    };
};

const

export class ObservableImpl<T> implements Observable<T> {

    private readonly _producer: FusedProducer<T>;

    constructor(producer: FusedProducer<T>) {
        this._producer = producer;
    }

    subscribe(consumer: Observer<T> = {}) {
        return {
            unsubscribe: this._producer(sink(consumer)) ?? noop
        }
    }

    map<T2>(mapper: (value: T) => T2) {
        return new ObservableImpl<T2>(({ next, ...rest }) =>
            this._producer({
                next: next !== noop ? (v: T) => next(mapper(v)) : noop,
                ...rest
            }));
    }

    tap({ next: tapNext, complete: tapComplete, error: tapError }: Observer<T>) {
        return new ObservableImpl<T>(({ next, complete, error }) =>
            this._producer({
                next: tapNext ? (v => (tapNext(v), next(v))) : next,
                complete: tapComplete ? (() => (tapComplete(), complete())) : complete,
                error: tapError ? (e => (tapError(e), error(e))) : error
            }));
    }

    filter: (predicate) => _observe(({ next, ...rest }) =>
        producer({
            next: next !== noop ? (v: T) => predicate(v) ? next(v) : void (0) : noop,
            ...rest
        })
    ),

until: predicate => _observe(({ next, complete, ...rest }) =>
    producer({
        next: v => predicate(v) ? complete() : next(v),
        complete,
        ...rest
    })
),

    while: predicate => _observe(({ next, complete, ...rest }) =>
    producer({
        next: v => predicate(v) ? next(v) : complete(),
        complete,
        ...rest
    })
),

    scan: <A>(...args: [AccumulatorFn<T, A>, A] | [AccumulatorFn<T, T>]) => _observe<T | A>(({ next, ...rest }) => {
        if (args.length === 1) {
            const [accumulator] = args;
            let _acc: T;
            let index = 0;
            return producer({
                next: next !== noop ? (v: T) => next(index++ === 0 ? _acc = v : _acc = accumulator(_acc, v)) : noop,
                ...rest
            });
        } else {
            const [accumulator, initial] = args;
            let _acc = initial;
            return producer({
                next: next !== noop ? (v: T) => next(_acc = accumulator(_acc, v)) : noop,
                ...rest
            });
        }
    }),

        reduce: <A>(...args: [AccumulatorFn<T, A>, A] | [AccumulatorFn<T, T>]) => _observe<T | A>(({ next, complete, error }) => {
            if (args.length === 1) {
                const [accumulator] = args;
                let _acc: T;
                let index = 0;
                return producer({
                    next: next !== noop ? (v: T) => {
                        _acc = index++ === 0 ? v : accumulator(_acc, v);
                    } : noop,
                    complete: () => {
                        if (index === 0) {
                            error(new Error("The sequence can't be empty"));
                        } else {
                            next(_acc);
                            complete();
                        }
                    },
                    error
                });
            } else {
                const [accumulator, initial] = args;
                let _acc = initial;
                return producer({
                    next: next !== noop ? (v: T) => {
                        _acc = accumulator(_acc, v);
                    } : noop,
                    complete: () => {
                        next(_acc);
                        complete();
                    },
                    error
                });
            }
        }),

            cat: (...seq) => _observe(({ next, complete: final, ...rest }) => {
                let cleanup: () => void;
                const len = seq.length;
                const complete = (i: number) => i < len ?
                    () => {
                        const subscription = seq[i].subscribe({ next, complete: complete(i + 1), ...rest });
                        cleanup = subscription.unsubscribe.bind(subscription);
                    } : final;

                cleanup = producer({ next, complete: complete(0), ...rest }) ?? noop;

                return () => cleanup();
            }),

                pipe: <U>(op: OperatorFn<T, U>) => op(_observe(producer)),

                    next: collect(
                        producer,
                        ({ next, complete, error }) => ({
                            next: v => (next(v), complete()),
                            complete: () => error(new Error("The sequence is empty")),
                            error
                        })
                    ),

                        collect: collect(
                            producer,
                            ({ next, complete, error }) => {
                                const data: T[] = [];
                                return {
                                    next: v => data.push(v),
                                    complete: () => (next(data), complete()),
                                    error
                                };
                            }
                        )
}